//***************************************************************************//
//* FRIDA: fast reliable interactive data analysis                          *//
//* (C) Joachim Wuttke 1990-, v2(C++) 2001-                                 *//
//* http://apps.jcns.fz-juelich.de/frida                                    *//
//***************************************************************************//

//! \file  expr.hpp
//! \brief Expressions are parsed into CRef, CContext, CResult, virtual CNode.

#include <boost/shared_ptr.hpp>

typedef boost::shared_ptr<class CNode>   PNode;
typedef boost::shared_ptr<class CVar>    PVar;
typedef boost::shared_ptr<class CRef>    PRef;
typedef boost::shared_ptr<class CRegRef> PRgr;


//! The context in which a CNode shall be evaluated.

class CContext {
 public:
    int k;                         //!< Index of file we are refering to
    int j;                         //!< Index of scan we are refering to
    int i;                         //!< Index of point we are refering to
    const vector<double> *vt;      //!< A vector we may refer to
    double vt_step;                //!< Step of equidistant vt, otherwise 0
    enum TDim { _1, _VI, _VT } dim;//!< Kind of requested output
    bool want_error;               //!< Are errors requested?
    int nv;                        //!< Number of requested vector elements

    CContext( int _k=-1, int _j=-1, int _i=-1 );

    void request_VT( const vector<double> *_vt );
    void request_VI( int _nv );

    string context_info() const;
};


//! The return-value container for tree evaluation.

class CResult {
 public:
    bool vectorial;    //!< Is it a vector?
    double r;          //!< Scalar value of result.
    double dr;         //!< Error of scalar result.
    vector<double> v;  //!< Resulting vector.
    vector<double> dv; //!< Errors of vectorial result.

    //! Creates meaningless defaut instance.
    CResult() : vectorial(false), dr(0) {}

    //! Sets scalar result without error.
    void set_r( double _r ) {
        vectorial = false; r = _r; }
    //! Prepares for vectorial result.
    void preset_v( int n, bool has_dv ) {
        vectorial = true; v.resize(n); dv.resize( has_dv ? n : 0 ); }
    //! Sets result vector, copying from argument.
    void set_v( const vector<double>& _v ) {
        vectorial = true; v = _v; }

    //! Returns result for given i.
    double to_r( int i ) const {
        return vectorial ? v[i] : r; }
    //! Returns error of result for given i.
    double to_dr( int i ) const {
        return vectorial ? ( dv.size() ? dv[i] : 0 ) : dr; }
    //! Is there a nonzero error?
    bool has_err() const {
        return vectorial ? dv.size() : dr; }

    string result_info() const;
};


//! A tuple of nodes that yield indices k,j,i.

class CRef {
 public:
    PNode tk; //!< A node yielding the workspace index k.
    PNode tj; //!< A node yielding the spectrum index j.
    PNode ti; //!< A node yielding the point index i.

    CRef( PNode _tk=PNode(), PNode _tj=PNode(), PNode _ti=PNode() ) 
        : tk(_tk), tj(_tj), ti(_ti) {}
    
    int get_k( const CContext& ctx ) const;
    int get_j( const CContext& ctx, int nj ) const;
    string ref_info() const;
};


//! Virtual base class for all kinds of nodes in an expression tree.

class CNode {
 public:
    virtual ~CNode() {}
    virtual void npar_exec( int *np ) const =0;
    //! Recursive: set *kd=true if a subnode is K dependent
    virtual void kdep_exec( bool *kd ) const {;} //!< default: do nothing

    int npar() const;
    bool k_dependent() const;
    virtual bool has_dummy() const =0; //!< does node depend on t ?
    virtual bool has_conv() const =0;  //!< does node depend on resolution ?

    //! Evaluate tree. Private. User calls pass through API implemented below.
    virtual void tree_val( CResult& ret, const CContext& ctx ) const =0;
    //! API for tree evaluation: compute one index.
    void tree_uival( int *ret, const CContext& ctx ) const;
    //! API for tree evaluation: compute one numeric value.
    void tree_point_val(
        double *ret, double *dret,
        int k=-1, int j=-1, int i=-1 ) const;
    //! API for tree evaluation: compute an array.
    void tree_vec_val( vector<double> *ret, vector<double> *dret,
                       int k=-1, int j=-1 ) const;

    virtual void set_coord( CCoord& ret, int k=-1 ) const =0;
    virtual string tree_info() const =0;  //!< introspection, for debugging
};
