//****************************************************************************//
//* FRIDA: fast reliable interactive data analysis                           *//
//* (C) Joachim Wuttke 1990-, v2(C++) 2001-                                  *//
//* http://apps.jcns.fz-juelich.de/frida                                     *//
//****************************************************************************//

//! \file  fit.cpp
//! \brief NCurveFit: fit a curve file to a data file

#include <iostream>
#include <string>

#include <omp.h>
#include <lmmin.h>
#include <readplus/ask.hpp>

#include "defs.hpp"
#include "olf.hpp"
#include "mem.hpp"
#include "slice.hpp"
#include "fit.hpp"
#include "expr.hpp"

bool allow_slow_conv = true;

//****************************************************************************//
//* Fit tuning parameters                                                    *//
//****************************************************************************//

namespace NCurveFit {
    double Epsilon = 1.e-14;
    double Factor=100.;
    int nCall=200;
    int verbosity=0;
    double maxtime = 3;
}

//! Set auxiliary parameters for the fit. Directly called from main program.

void NCurveFit::setFitTuningPars( string which )
{
    if      ( which=="?" )
        printf( "Fit setup:\n   nCall=%d\n   Verbosity=%d\n   Factor=%g\n"
                "   Eps=%g\n",
                nCall, verbosity, Factor, Epsilon );
    else if ( which=="v" )
        verbosity = iask( "Verbosity", verbosity );
    else if ( which=="c" )
        nCall = iask( "Max no of Levenberg-Marquardt steps", nCall );
    else if ( which=="f" )
        Factor = dask( "NTuneFit::Factor", Factor );
    else if ( which=="e" )
        Epsilon = dask( "NTuneFit::Epsilon", Epsilon );
    else if ( which=="t" )
        maxtime = dask( "Max duration of fits (in s)", maxtime );
}

//****************************************************************************//
//* Least-squares fitting: call back                                         *//
//****************************************************************************//

//! Data transfer between frida and lm_minimize callback functions.

typedef struct {
    POld fd;
    POlc fc;
    int k;
    int j;
    double timeout;
} FitDatTyp;

//! Callback routine for use in lm_minimize: function evaluation.

static void fitEvaluate( const double* par, int m_dat, const void *data, 
                         double* fvec, int *userbreak )
// arguments (required by lmmin):
//    par       r/w  function parameters, to be optimized
//    m_dat     r    number of data points
//    data      r    pointer to arbitrary data, here defined as FitDatTyp
//    fvec      w    vector of residues, to be minimized
//    userbreak w    can be use to terminate the minimization
{
    try {
        FitDatTyp *mydata;
        mydata = (FitDatTyp*) data;
        if( mydata->timeout && time(NULL) > mydata->timeout )
            throw "reached fit time out";
        POlc fc = mydata->fc;
        POld fd = mydata->fd;
        COlc::TWgt wt = fc->weighing;
        PCurve c = fc->VC(mydata->j);
        
        int ip = 0;
        for( int i=0; i<fc->nP; ++i )
            if ( !(fc->VC(mydata->j)->fixed[i]) )
                c->P[i] = par[ip++];
        
        const PSpec s = fd->VS(mydata->j);
        vector<double> v, dv;
        vector<double> *pdv = (wt==COlc::_VAR || wt==COlc::_VARC) ? &dv : NULL;

        fc->curve_val_vec( &v, pdv, s->x, mydata->k, mydata->j );

        if ( ( wt==COlc::_VAR || wt==COlc::_VARD ) && s->dy.size()!=m_dat )
            throw "missing data error bars for weighing";
        if ( pdv && pdv->size()!=m_dat )
            throw "BUG in fitEvaluate: curve_val_vec did not resize dv";

        for ( int i=0; i<m_dat; i++) {
            if( v[i]==INFINITY ) // TEMPORARY UNTIL exceptions can be
                                   // thrown everywhere
                throw "function evaluation return INFINITY";

            if        ( wt==COlc::_LIN ) {
                fvec[i] = s->y[i] - v[i];
            } else if ( wt==COlc::_LOG ) {
                if ( s->y[i] < 1e-99 ) {
                    throw "cannot weigh with log y: spectrum " +
                        S(mydata->j) + " channel " + S(i) +
                        " has y=" + S(s->y[i]);
                } else if ( v[i] < 1e-99 ) {
                    fvec[i] = 99; // giant penalty
                } else {
                    fvec[i] = log( s->y[i] / v[i] );
                }
            } else if ( wt==COlc::_VAR || wt==COlc::_VARC || wt==COlc::_VARD ) {
                double dr;
                
                if     ( wt==COlc::_VAR )
                    dr = sqrt( SQR(s->dy[i]) + SQR((*pdv)[i]) );
                else if( wt==COlc::_VARC )
                    dr = (*pdv)[i];
                else if( wt==COlc::_VARD )
                    dr = s->dy[i];
                else
                    throw "BUG: unforeseen case";
                if ( dr < 1e-99 )
                    throw "cannot weigh with dy: spectrum " + S(mydata->j) +
                        " channel " + S(i) + " has y=" + S(s->y[i]) +
                        ", dy=" + S(s->dy[i]) + "; theory yields y=" + S(v[i]) +
                        ", dy=" + (pdv ? S((*pdv)[i]) : S("n/a"));
                        
                fvec[i] = ( s->y[i] - v[i] ) / dr;
            }
        }
    } catch ( string& s ) {
        cerr << "\n" << s << "\n";
        *userbreak = -1;
    } catch ( const char* s ) {
        cerr << "\n" << s << "\n";
        *userbreak = -1;
    } catch ( ... ) {
        cerr << "BUG: fitEvaluate catched invalid exception\n";
        *userbreak = -1;
    }
}

//****************************************************************************//
//* Least-squares fitting: main                                              *//
//****************************************************************************//

//! Fit selected curve files to the data files they refer to.

void NCurveFit::fit( bool _allow_slow_conv )
{
    lm_control_struct control = lm_control_double;;
    allow_slow_conv = _allow_slow_conv;

    NOlm::mem_backup_store();
    NOlm::IterateC fiter;

    while ( POlc fc = fiter() ) {
        POld fd = NOlm::mem_get_D(fc->kd);
        if ( fc->nJ() != fd->nJ() )
            throw "data file and curve file have different number of spectra";
        if ( !fc->nJ() )
            throw "BUG: empty curve and data file";
        if ( ( fc->weighing==COlc::_VAR || fc->weighing==COlc::_VARD ) &&
             !fd->VS(0)->dy.size() )
            throw "Data have no error bars, cannot apply chosen fit weights";

        if( fc->range_T ) {
            // overwrite fd with data that obye the range restriction:
            POld fdr( fd->new_old() );
            for ( int j=0; j<fc->nJ(); j++ ) {
                PSpec ein = fd->VS(j);
                PSpec eout = PSpec( new CSpec );
                eout->copy_z_base( ein );
                vector<double> range( ein->size() );
                fc->range_T->tree_vec_val( &range, 0, fc->kd, j );
                bool with_dy = ein->dy.size();
                for ( int i=0; i<ein->size(); i++ ) {
                    if ( range[i] ) {
                        if( with_dy )
                            eout->push_xyd( ein->x[i], ein->y[i], ein->dy[i] );
                        else
                            eout->push_xy( ein->x[i], ein->y[i] );
                    }
                }
                fdr->V.push_back( eout );
            }
            fd = fdr;
        }
        

        int np = fc->nP;

        control.epsilon = Epsilon;
        control.stepbound = Factor;
        control.patience = nCall;
        control.verbosity = verbosity;

#pragma omp parallel for
        for ( int j=0; j<fc->nJ(); j++ ) {
            try {
                if ( fc->V[j]->frozen )
                    continue;
                FitDatTyp data;
                data.fc = fc;
                data.fd = fd;
                data.k = fiter.k();
                data.j = j;
                PSpec D = fd->VS(j);
                PCurve C = fc->VC(j);
                int nd = D->size();
                if ( !nd ){
                    printf( "Empty data set in spectrum %i\n", j );
                    continue;
                }

                // handle fixed parameters:
                int npfree = 0;
                for ( int ip=0; ip<np; ++ip )
                    if( !fc->VC(j)->fixed[ip] )
                        ++npfree;
                if ( npfree<=0 )
                    continue;
                vector<double> Par(npfree);
                int ipf=0;
                for ( int ip=0; ip<np; ip++ ) {
                    if ( !fc->VC(j)->fixed[ip] )
                        Par[ipf++] = C->P[ip];
                }
                                        
                // Levenberg-Marquardt routine from library lmfit:

                data.timeout = time(NULL) + omp_get_num_threads()*maxtime;

                lm_status_struct status;
                lmmin( npfree, &(Par[0]), nd, &data,
                       fitEvaluate, &control, &status );

                if( status.outcome==11 )
                    throw "exception in fit function";

                // Fit quality metrics Q:

                // Q: status code
                C->fitOutcome = (double) status.outcome;

                // Q: coefficient of determination R^2 (for linear regression)
                double ysum1 = 0, ysum2 = 0;
                for( int i=0; i<nd; ++i ){
                    ysum1 += D->y[i];
                    ysum2 += SQR( D->y[i] );
                }
                double ydev2 = ysum2 - SQR( ysum1 )/nd; // = nd * <(y-<y>)^2>
                C->fitR2 = ydev2!=0 ? 1 - SQR(status.fnorm) / ydev2 : -1;
                        
                // Q: mean deviation (chi^2 if fvec is weighed with 1/dy)
                int nfreedom = nd - npfree - 1;
                if( nfreedom<1 ) nfreedom = 1;
                C->fitChi2 = SQR(status.fnorm) / nfreedom;

                {
                    // print results:
                    if( fiter.size()>1 )
                        printf( "%3d", data.k );
                    printf( "%3d", j );
                    for ( int i=0; i<npfree; i++ )
                        printf(" %12g", Par[i]);
                    printf( " (%3d->%9.7g)",
                            status.nfev, C->fitChi2 );
                    printf( "%s\n", lm_shortmsg[ status.outcome ] );
                }
            } catch (...) {
                printf( "Caught unexpected exception"
                        " while fitting spectrum %i\n", j );
            } // try
        } // j
    } // k
}
