//***************************************************************************//
//* FRIDA: fast reliable interactive data analysis                          *//
//* (C) Joachim Wuttke 2001-                                                *//
//* http://apps.jcns.fz-juelich.de/frida                                    *//
//***************************************************************************//

//! \file  func.cpp
//! \brief Collection NFunctions of function wrappers CFunc.


#include <iostream>
#include <vector>
#include <map>
#include <gsl/gsl_sf.h>

#include <trivia/rng.hpp>
#include <kww.h>
#include <cerf.h>

#include "defs.hpp"
#include "func.hpp"
#include "coord.hpp"

const double PI    = M_PI;    //4*atan(1.);
const double TWOPI = 2*M_PI; //8*atan(1.);

static triv::RandomNumberGenerator *rng = NULL;


//**************************************************************************//
//*   Functions of one argument                                            *//
//**************************************************************************//

double func_not(double v) { return !v; }
double func_neg(double v) { return -v; }
void deri_neg(double& r, double& dr, double v, double dv)
{ r=-v; dr=dv; }

double func_ln(double v) { return v>0 ? log(v) : 0; }
void deri_ln(double& r, double& dr, double v, double dv)
{ if(v>0) { r=log(v); dr=dv/v; } else { r=0; dr=0; } }

double func_lg(double v) { return v>0 ? log10(v) : 0; }
void deri_lg(double& r, double& dr, double v, double dv)
{ if(v>0) { r=log10(v); dr=dv/v/log(10); } else { r=0; dr=0; } }

double func_sqrt(double v) { return v>=0 ? sqrt(v) : 0; }
void deri_sqrt(double& r, double& dr, double v, double dv)
{ if(v>0) { r=sqrt(v); dr=dv/2/r; } else { r=0; dr=0; } } 

double func_abs(double v) { return fabs(v); }
void deri_abs(double& r, double& dr, double v, double dv)
{ r=fabs(v); dr=dv; }

double func_sign(double v) {
    if      ( v<0 )
        return -1;
    else if ( v>0 )
        return 1;
    else
        return 0;
}
void deri_sign(double& r, double& dr, double v, double dv)
{ r=func_sign(v); dr=dv; }

double func_exp(double v) { return exp(v); }
void deri_exp(double& r, double& dr, double v, double dv)
{ r=exp(v); dr=dv*r; }

double func_sin(double v) { return sin(v); }
double func_cos(double v) { return cos(v); }
double func_tan(double v) { return tan(v); }
double func_cot(double v) { double t=tan(v); return t==0?0:1/t;}

double func_sind(double v) { return sin(v*TWOPI/360.); }
double func_cosd(double v) { return cos(v*TWOPI/360.); }
double func_tand(double v) { return tan(v*TWOPI/360.); }
double func_cotd(double v) { double t=tan(v*TWOPI/360.);  return t==0?0:1/t; }

double func_asin(double v) { return (-1<=v && v<=1) ? asin(v) : 0; }
double func_acos(double v) { return (-1<=v && v<=1) ? acos(v) : 0; }
double func_atan(double v) { return atan(v); }
double func_acot(double v) { return v!=0 ? atan(1/v) : TWOPI/4; }

double func_asind(double v) { return (-1<=v && v<=1) ? 360./TWOPI*asin(v) : 0; }
double func_acosd(double v) { return (-1<=v && v<=1) ? 360./TWOPI*acos(v) : 0; }
double func_atand(double v) { return 360./TWOPI*atan(v); }
double func_acotd(double v) { return v!=0 ? 360./TWOPI*atan(1/v) : 90; }

double func_sinh(double v) { return sinh(v); }
double func_cosh(double v) { return cosh(v); }
double func_tanh(double v) { return tanh(v); }
double func_coth(double v) { double t=tanh(v); return t==0?0:1/t;}

double func_j0(double v) { return v==0 ? 1 : sin(v)/v; }
double func_j1(double v) { return v==0 ? 1 : (sin(v)/v-cos(v))/v; }

double func_fac(double v) { return v>-.5 ? gsl_sf_gamma(round(v+1)) : 0; } 
double func_cata(double v) {
    if ( v<=-.5 )
        return 0;
    double x = round(v);
    return exp( gsl_sf_lngamma( 2*x+1 ) - gsl_sf_lngamma( x+1 ) -
                gsl_sf_lngamma( x+2 ) ); }

double func_gamma(double v) { return v>0 ? gsl_sf_gamma(v) : 0; } 
double func_lgamma(double v) { return v>0 ? gsl_sf_lngamma(v) : 0; } 
// could be extended to non-integer negative arguments...

double func_debye1(double v) { return v>=0 ? gsl_sf_debye_1(v) : 0; }
double func_debye3(double v) { return v>=0 ? gsl_sf_debye_3(v) : 0; }
double func_debye_msd(double v) { // mean squared displacement
                                  // without prefactors
                                  // without zero temperature oscillations
    return v>1e-2 ? 2*v*gsl_sf_debye_1(1/v) : 0; }
double func_debye_ui(double v) { // internal energy
    return v>=0 ? 3*v*gsl_sf_debye_3(1/v) : 0; }
double func_debye_cv(double v) { // heat capacity
    double eps=1e-12;
    return v>=eps ? (func_debye_ui(v+eps)-func_debye_ui(v-eps))/(2*eps) : 0; }

double func_erfP(double v) { return 1-gsl_sf_erf_Q(v); }
double func_erfQ(double v) { return gsl_sf_erf_Q(v); } 
double func_erf(double v) { return gsl_sf_erf(v); } 
double func_erfc(double v) { return gsl_sf_erfc(v); } 

double func_dawson(double v) { return gsl_sf_dawson(v); }

double func_sinc(double v) { return v==0 ? 1 : sin(v)/v; }

double func_ceil(double v)  { return ceil(v); }
double func_floor(double v) { return floor(v); }
double func_nint(double v)  { return round(v); }

// Ad hoc for uses in physics:

double func_lndiehl( double a ) {
    // ln( \int dw gauss(w,gausswid) cauchy(w,cauchywid) / gauss(0,gausswid) )
    // argumnt a = cauchywid/gausswid
    return SQR(a)/2 + gsl_sf_log_erfc( a/sqrt(2) );
}

double func_diehl( double a ) {
    // \int dw gauss(w,gausswid) cauchy(w,cauchywid) / gauss(0,gausswid)
    // argumnt a = cauchywid/gausswid
    return exp( func_lndiehl(a) );
}


//**************************************************************************//
//*   Functions of two arguments                                           *//
//**************************************************************************//

double func_pm(double v, double a) { return v; }
void deri_pm(double& r, double& dr, double v, double dv, double a, double da)
{ r=v; dr=a; }

double func_add(double v, double a) { return v+a; }
void deri_add(double& r, double& dr, double v, double dv, double a, double da)
{ r=v+a; dr=sqrt(dv*dv+da*da); }

double func_sub(double v, double a) { return v-a; }
void deri_sub(double& r, double& dr, double v, double dv, double a, double da)
{ r=v-a; dr=sqrt(dv*dv+da*da); }

double func_mul(double v, double a) { return v*a; }
void deri_mul(double& r, double& dr, double v, double dv, double a, double da)
{ r=v*a; dr=sqrt(SQR(v*da)+SQR(a*dv)); }

double func_div(double v, double a) { return a!=0 ? v/a : 0; }
void deri_div(double& r, double& dr, double v, double dv, double a, double da)
{ if(a==0) { r=0; dr=0; } else { r=v/a; dr=sqrt(SQR(dv/a)+SQR(v*da/a/a)); } }

double func_pow(double v, double a)
{ return ( v>0 || (v==0&&a>0) || a==(int)a ) ? pow(v,a) : 0; }
void deri_pow(double& r, double& dr, double v, double dv, double a, double da)
{
    if ( v>0 || (v==0&&a>0) || a==(int)a ) {
        r = pow(v,a); dr = sqrt( SQR(a*r/v) + SQR(log(v)*a) );
    } else {
        r = 0; dr = 0;
    }
}

double func_eq (double v, double a) { return v == a; }
double func_ne (double v, double a) { return v != a; }
double func_gt (double v, double a) { return v > a; }
double func_ge (double v, double a) { return v >= a; }
double func_lt (double v, double a) { return v < a; }
double func_le (double v, double a) { return v <= a; }
double func_and(double v, double a) { return v && a; }
double func_or (double v, double a) { return v || a; }
double func_xor(double v, double a) { return (bool) v ^ (bool) a; }

double func_idiv(double v, double a) { return a!=0 ? ((int)v)/((int)a) : 0; }
double func_mod(double v, double a) { return fmod(v,a); }
double func_over(double v, double a) {
    return  ( 0<=v && 0<=a && a<=v && v <= GSL_SF_GAMMA_XMAX ) ?
        gsl_sf_gamma(v+1)/gsl_sf_gamma(a+1)/gsl_sf_gamma(v-a+1) :
        0; }

double func_min (double v, double a) { return v<a ? v : a; }
double func_max (double v, double a) { return v>a ? v : a; }

double func_gauss (double x, double s) {
    return s==0 ? 0 : exp(-x*x/(2*s*s))/sqrt(TWOPI*s*s); }
double func_gaussnn (double x, double s) { // Gauss not normalized
    return s==0 ? 0 : exp(-x*x/(2*s*s)); }
double func_cauchy (double x, double s) {
    return s==0 ? 0 : fabs(s)/PI/(x*x+s*s); }

double func_ran (double v, double a) { 
    // a random number between v and a :
    if (!rng)
        rng = new triv::RandomNumberGenerator();
    return rng->uniform( func_min(v,a), func_max(v,a) ); }

// Havriliak-Negami: unchecked formulae from disreputable source:

double aux_phi_havneg (double v, double a){
    return atan( pow(v,a)*sin(M_PI*a/2)/(1+pow(v,a)*cos(M_PI*a/2)) ); }

double func_re_havneg (double v, double a, double g) {
    return pow( 1 + 2*pow(v,a)*cos(M_PI*a/2)+pow(v,2*a), -g/2 ) *
        cos( g*aux_phi_havneg( v, a ) ); }

double func_im_havneg (double v, double a, double g) {
    return pow( 1 + 2*pow(v,a)*cos(M_PI*a/2)+pow(v,2*a), -g/2 ) *
        sin( g*aux_phi_havneg( v, a ) ); }


//**************************************************************************//
//*   Functions of three arguments                                         *//
//**************************************************************************//

double func_cond(double v, double a, double b) { return v ? a : b; }
void deri_cond(double& r, double& dr, double v, double dv,
               double a, double da, double b, double db)
{ if(v) { r=a; dr=da; } else { r=b; dr=db; } }

double func_q4w( double hbarw, double Ei, double twotheta) { 
    // calculate Q in Angstroem-1 when hbar*w and E_i are given in meV
    if( hbarw>Ei ) return -1;
    return sqrt( (1/2.07212349) *
                 ( 2*Ei-hbarw-2*sqrt(Ei*(Ei-hbarw))*cos(twotheta*TWOPI/360) ) );
}

double func_cauchy2 (double x, double p, double w) {
    return w==0 ? 0 : w/PI*(1.0/(SQR(x-p)+SQR(w))+1.0/(SQR(x+p)+SQR(w)));
}

double func_kwwc( double v, double t, double b ){
    if( b<0.1 || b>2.0 ) return 0;
    return t/PI*kwwc( v*t, b ); }
double func_kwws( double v, double t, double b ){
    if( b<0.1 || b>2.0 ) return 0;
    return t/PI*kwws( v*t, b ); }
double func_kwwp( double v, double t, double b ){
    if( b<0.1 || b>2.0 ) return 0;
    return 1/PI*kwwp( v*t, b ); }

double func_voigt( double v, double s, double g ){
    return (s||g) ? voigt(v,s,g) : 0;
}

double func_zorn( double v, double vavg, double s ) {
    if( s<=0 || s>=1 ) return 0;
    return (1-s)*(v+s*vavg*vavg/(1-s*vavg)); }

double func_zorn_gauss( double q, double u2, double s ) {
    if( s<=0 || s>=1 ) return 0;
    double qmax = 4*PI/6.27; // Si111
    double v = exp(-u2*SQR(q));
    double vavg = (1-exp(-u2*SQR(qmax))) / (u2*SQR(qmax));
    return (1-s)*(v+s*vavg*vavg/(1-s*vavg)); }

double func_rrdm( double wred, double act0, double relwidth ) {
    // Rotational rate distribution model [review: Colmenero 2005]
    //   Gaussian distribution of Barriers, Arrhenius law, rotational jumps.
    // Dimensionless:
    //   wred = w/Omega_0,
    //   act0 = mean{E_A}/(k_B T),
    //   relwidth = stdv{E_A} / mean{E_A};
    //   multiply result by T/Omega_0.
    const int N=101;
    double actsig = act0*relwidth;
    double actrange = 3*actsig;
    if ( act0-actrange <= 0 )
        return 0;
    double da = actrange/((N-1)/2);
    double ret = 0;
    for( int i=0; i<N; ++i ) {
        double a = act0 + da*(i-(N-1)/2);
        ret += exp( -SQR((a-act0)/actsig)/2 ) * func_cauchy( wred, exp(-a) );
    }
    return ret * da / sqrt(2*PI) / actsig;
}

double func_rotdiff( double w, double tau, double qb )
// Rotational diffusion spectrum
{
    double ret = 0;
    if( qb<=0 )
        return 0;
    for( int l=1; l<=12; ++l ){
        ret += (2*l+1) * pow( gsl_sf_bessel_jl( l, qb ), 2 ) *
            l*(l+1)*tau/PI/(SQR(w*tau)+SQR(l*(l+1)));
    }
    return ret;
}


//**************************************************************************//
//*   Coordinate concatenation                                             *//
//**************************************************************************//

CCoord CFunc1::coord( CCoord *co1 ) const
{
    CCoord co = *co1;
    if (txt=="0-") {
        co.name = "-" + co1->name;
    } else if (txt=="ln") {
        co.name = "ln " + co1->name;
    } else if (txt=="lg") {
        co.name = "lg " + co1->name;
    } else if (txt=="sqrt") {
        co.name = "sqrt (" + co1->name + ")";
        co.unit = co1->unit=="" ? "" : "(" + co1->unit + ")^1/2";
    } else {
        co.name = txt + "(" + co1->name + ")";        
    }
    return co;
}

CCoord CFunc2::coord( class CCoord *co1, class CCoord *co2 ) const
{
    CCoord co = *co1;

    if ( txt=="+" || txt=="-" ) {
        if( co1->name==co2->name ){
            if( co1->unit!=co2->unit )
                throw "same coordinate name and different units" 
                    " in additive operation";
            co = *co1;
        } else {
            co.name = co1->name + txt + co2->name;
            if ( co1->unit=="" )
                co.unit = "";
            else if ( co2->unit!="" && co2->unit!=co1->unit) {
                cout << "? different units " << co1->unit << " and " <<
                    co2->unit << " in additive operation\n";
                co.unit = "";
            } else
                co.unit = co1->unit;
        }
    } else if ( txt=="*" || txt=="/" ) {
        co.name = co1->name + txt + co2->name;
        if     ( co1->unit!="" && co2->unit!="" )
            co.unit = co1->unit + txt + co2->unit;
        else if( co1->unit!="" )
            co.unit = co1->unit;
        else if( co2->unit!="" )
            co.unit = txt=="*" ? co2->unit : "1/(" + co2->unit + ")";
        else
            co.unit = "";
    } else if ( txt=="^" ) {
        if (co2->unit != "") {
            cout << "? exponent is not dimensionless\n";
        }
        co.name = co1->name + "^" + co2->name;
        co.unit = co1->unit=="" ? "" : ( co1->unit + "^<val>" );
    } else {
        co.name = txt + "(" + co1->name + "," + co2->name + ")";
        co.unit = "";
    }
    return co;
}

CCoord CFunc3::coord( class CCoord *co1, class CCoord *co2, class CCoord *co3 )
    const
{
    CCoord co;
    co.name = txt + "(" + co1->name + "," + co2->name + "," + co3->name + ")";
    co.unit = "";
    return co;
}


//**************************************************************************//
//*   Function registration and retrieval                                  *//
//**************************************************************************//

namespace NFunctions { // internals:
    map<string,const CFunc*> fmap; // unsorted hash, for expression evaluation
    vector<const CFunc*> flist;    // sorted array, for help text
}

const CFunc* NFunctions::find( string key )
{
    auto pos = fmap.find(key);
    return pos == fmap.end() ? NULL : pos->second;
}

void NFunctions::display_functions()
{
    // this implementation works only, if flist is strictly ordered:
    // first operators by precedence, then functions.

    cout << "Operators by precedence:\n";
    int prec = -1;
    for( int i=0; i<flist.size(); ++i ){
        const CFunc *f;
        f = flist[i];
        if( f->prec ) { // operators
            if( f->prec!=prec )
                cout << "\n ";
            cout << " " << f->txt;
        } else { // functions
            if( f->prec!=prec )
                cout << "\nBuilt-in functions:\n";
            cout << "  " << f->txt << f->com << "\n";
            }
        prec = f->prec;
    }
}

void CFunc::register_me() const
{
    NFunctions::fmap.insert( make_pair( txt, this ) );
    NFunctions::flist.push_back( this );
}

void NFunctions::initialize(void)
{
    // operators by precedence (as in xax_yacc.ypp):
    (new CFunc2( "+-",  func_pm,   deri_pm,    1   ))->register_me();
    (new CFunc2( "^",   func_pow,  deri_pow,   2   ))->register_me();
    (new CFunc1( "!",   func_not,  0,          3   ))->register_me();
    (new CFunc1( "neg", func_neg,  deri_neg,   4   ))->register_me();
    (new CFunc2( "over",func_over, 0,          5   ))->register_me();
    (new CFunc2( "*",   func_mul,  deri_mul,   6   ))->register_me();
    (new CFunc2( "/",   func_div,  deri_div,   6   ))->register_me();
    (new CFunc2( "mod", func_mod,  0,          6   ))->register_me();
    (new CFunc2( "div", func_idiv, 0,          6   ))->register_me();
    (new CFunc2( "+",   func_add,  deri_add,   7   ))->register_me();
    (new CFunc2( "-",   func_sub,  deri_sub,   7   ))->register_me();
    (new CFunc2( ">",   func_gt,   0,          8   ))->register_me();
    (new CFunc2( ">=",  func_ge,   0,          8   ))->register_me();
    (new CFunc2( "<",   func_lt,   0,          8   ))->register_me();
    (new CFunc2( "<=",  func_le,   0,          8   ))->register_me();
    (new CFunc2( "==",  func_eq,   0,          9   ))->register_me();
    (new CFunc2( "!=",  func_ne,   0,          9   ))->register_me();
    (new CFunc2( "xor", func_xor,  0,         10   ))->register_me();
    (new CFunc2( "&&",  func_and,  0,         11   ))->register_me();
    (new CFunc2( "||",  func_or ,  0,         12   ))->register_me();
    (new CFunc3( "?:",  func_cond, deri_cond, 13 ))->register_me();

    // f( 1 arg )
    (new CFunc1( "ln",   func_ln,    deri_ln, 0,
           "(x): natural logarithm of x, or 0 if x<=0" ))->register_me();
    (new CFunc1( "lg",   func_lg,    deri_lg, 0,
           "(x): decadic logarithm of x, or 0 if x<=0" ))->register_me();
    (new CFunc1( "sqrt", func_sqrt,  deri_sqrt, 0,
           "(x): square root of x, or 0 if x<0" ))->register_me();
    (new CFunc1( "abs",  func_abs,   deri_abs, 0,
           "(x): absolute value of x" ))->register_me();
    (new CFunc1( "abs",  func_sign,   deri_sign, 0,
           "(x): sign of x" ))->register_me();
    (new CFunc1( "exp",  func_exp,   deri_exp, 0,
           "(x): exponential function of x" ))->register_me();
    (new CFunc1( "sin",  func_sin, 0, 0,
           "(x): sine of x (where x in radian)" ))->register_me();
    (new CFunc1( "cos",  func_cos, 0, 0,
           "(x): cosine of x (where x in radian)" ))->register_me();
    (new CFunc1( "tan",  func_tan, 0, 0,
           "(x): tangent of x (where x in radian)" ))->register_me();
    (new CFunc1( "cot",  func_cot, 0, 0,
           "(x): cotangent of x (where x in radian)" ))->register_me();
    (new CFunc1( "sind",  func_sind, 0, 0,
           "(x): sine of x (where x in degrees)" ))->register_me();
    (new CFunc1( "cosd",  func_cosd, 0, 0,
           "(x): cosine of x (where x in degrees)" ))->register_me();
    (new CFunc1( "tand",  func_tand, 0, 0,
           "(x): tangent of x (where x in degrees)" ))->register_me();
    (new CFunc1( "cotd",  func_cotd, 0, 0,
           "(x): cotangent of x (where x in degrees)" ))->register_me();
    (new CFunc1( "asin", func_asin, 0, 0,
           "(x): arc sine of x (result is in radian; return 0 if |x|>1)"
        ))->register_me();
    (new CFunc1( "acos", func_acos, 0, 0,
           "(x): arc cosine of x (result is in radian; return 0 if |x|>1)"
        ))->register_me();
    (new CFunc1( "atan", func_atan, 0, 0,
           "(x): arc tangent of x (result is in radian)" ))->register_me();
    (new CFunc1( "acot", func_acot, 0, 0,
           "(x): arc cotangent of x (result is in radian)" ))->register_me();
    (new CFunc1( "asind", func_asind, 0, 0,
           "(x): arc sine of x (result is in degrees; return 0 if |x|>1)"
        ))->register_me();
    (new CFunc1( "acosd", func_acosd, 0, 0,
           "(x): arc cosine of x (result is in degrees; return 0 if |x|>1)"
        ))->register_me();
    (new CFunc1( "atand", func_atand, 0, 0,
           "(x): arc tangent of x (result is in degrees)" ))->register_me();
    (new CFunc1( "acotd", func_acotd, 0, 0,
           "(x): arc cotangent of x (result is in degrees)" ))->register_me();
    (new CFunc1( "sinh", func_sinh, 0, 0,
           "(x): hyperbolic sine of x" ))->register_me();
    (new CFunc1( "cosh", func_cosh, 0, 0,
           "(x): hyperbolic cosine of x" ))->register_me();
    (new CFunc1( "tanh", func_tanh, 0, 0,
           "(x): hyperbolic tangent of x" ))->register_me();
    (new CFunc1( "coth", func_coth, 0, 0,
           "(x): hyperbolic tangent of x" ))->register_me();
    (new CFunc1( "j0", func_j0, 0, 0,
           "(x): spherical Bessel function x" ))->register_me();
    (new CFunc1( "j1", func_j1, 0, 0,
           "(x): spherical Bessel function x" ))->register_me();
    (new CFunc1( "fac",func_fac, 0, 0,
           "(x): factorial of nearest integer of x" ))->register_me();
    (new CFunc1( "cata",func_cata, 0, 0,
           "(x): Catalan number of nearest integer of x" ))->register_me();
    (new CFunc1( "gamma",func_gamma, 0, 0,
           "(x): gamma function of x (i.e. factorial of x-1)" ))->register_me();
    (new CFunc1( "lgamma",func_lgamma, 0, 0,
           "(x): ln of gamma of x" ))->register_me();
    (new CFunc1( "debye1",func_debye1, 0, 0,
           "(x): Debye function D1 of x" ))->register_me();
    (new CFunc1( "debye3",func_debye3, 0, 0,
           "(x): Debye function D3 of x" ))->register_me();
    (new CFunc1( "debyeu2",func_debye_msd, 0, 0,
           "(x): Debye mean squared displacement's temperature dependence"
        ))->register_me();
    (new CFunc1( "debyeui",func_debye_ui, 0, 0,
           "(x): Debye internal energy" ))->register_me();
    (new CFunc1( "debyecv",func_debye_cv, 0, 0,
           "(x): Debye heat capacity" ))->register_me();
    (new CFunc1( "erfP", func_erfP, 0, 0,
           "(x): ?" ))->register_me();
    (new CFunc1( "erfQ", func_erfQ , 0, 0,
           "(x): ?"  ))->register_me();
    (new CFunc1( "erf",  func_erf, 0, 0,
           "(x): error function of x" ))->register_me();
    (new CFunc1( "erfc", func_erfc, 0, 0,
           "(x): complementary error function of x" ))->register_me();
    (new CFunc1( "dawson", func_dawson, 0, 0,
           "(x): Dawson function of x" ))->register_me();
    (new CFunc1( "sinc", func_sinc, 0, 0,
           "(x): sinus cardinalis, sin(x)/x" ))->register_me();
    (new CFunc1( "ceil", func_ceil, 0, 0,
           "(x): the smallest integer i with i>=x" ))->register_me();
    (new CFunc1( "floor",func_floor, 0, 0,
           "(x): the largest integer i with i<=x" ))->register_me();
    (new CFunc1( "nint", func_nint, 0, 0,
           "(x): the integer nearest to x" ))->register_me();

    (new CFunc1( "diehl", func_diehl, 0, 0,
           "(cauchywid/gausswid): normalized convolution gauss(*)cauchy"
        ))->register_me();
    (new CFunc1( "lndiehl", func_lndiehl, 0, 0,
           "(cauchywid/gausswid): ln of normalized convolution gauss(*)cauchy"
        ))->register_me();


    // f(2 args)
    (new CFunc2( "min2",  func_min, 0, 0,
           "(x,y): the smaller of the two arguments x and y" ))->register_me();
    (new CFunc2( "max2",  func_max, 0, 0,
           "(x,y): the smaller of the two arguments x and y" ))->register_me();
    (new CFunc2( "ran",  func_ran, 0, 0,
           "(x,y): a random number between x and y" ))->register_me();

    (new CFunc2( "gauss",   func_gauss, 0, 0,
           "(x,s): the normalized Gaussian exp(-x^2/2/s^2)/sqrt(2 pi)/s"
        ))->register_me();
    (new CFunc2( "gnn",     func_gaussnn, 0, 0,
           "(x,s): the unnormalized Gaussian exp(-x^2/2/s^2)" ))->register_me();
    (new CFunc2( "cauchy",  func_cauchy, 0, 0,
           "(x,w): the Cauchy-Lorentz function ?/(x^2+w^2)" ))->register_me();


    // f(3 args)
    (new CFunc3( "rehavneg", func_re_havneg, 0, 0,
           "(x,y,z): real part of the Havriliak-Negami function"
        ))->register_me();
    (new CFunc3( "imhavneg", func_im_havneg, 0, 0,
           "(x,y,z): imaginary part of the Havriliak-Negami function"
        ))->register_me();
    (new CFunc3( "q4w",      func_q4w, 0, 0,
           "(x,y,z): ?" ))->register_me();
    (new CFunc3( "cauchy2",  func_cauchy2, 0, 0,
           "(x,y,z): ?" ))->register_me();
    (new CFunc3( "kwwc", func_kwwc, 0, 0,
           "(w,tau,b): Fourier cosine transform (t->w) of exp((t/tau)^b)"
        ))->register_me();
    (new CFunc3( "kwws", func_kwws, 0, 0,
           "(w,tau,b): Fourier sine transform (t->w) of exp((t/tau)^b)"
        ))->register_me();
    (new CFunc3( "kwwp", func_kwwp, 0, 0,
           "(w,tau,b): primitive of "
           "Fourier cosine transform (t->w) of exp((t/tau)^b)"
        ))->register_me();
    (new CFunc3( "voigt", func_voigt, 0, 0,
           "(x,sigma,gamma): convolution of Gaussian(x,sigma) and "
           "Lorentzian(x,gamma)"
        ))->register_me();
    (new CFunc3( "zorn", func_zorn, 0, 0,
           "(I,<I>,s): Zorn's multiple-scattering corrected elastic intensity"
        ))->register_me();
    (new CFunc3( "zorn2", func_zorn_gauss, 0, 0,
           "(q,<u^2>,s): Zorn's multiple-scattering corrected Gaussian "
           "elastic intensity for Si111" ))->register_me();
    (new CFunc3( "rrdm", func_rrdm, 0, 0,
           "(w*t0,EA_mean/T,EA_stdv/EA_mean: rotational rate distribution model"
        ))->register_me();
    (new CFunc3( "rotdiff", func_rotdiff, 0, 0,
           "(w,tau,qb: rotational diffusion spectrum)"))->register_me();
}
