//***************************************************************************//
//* FRIDA: fast reliable interactive data analysis                          *//
//* (C) Joachim Wuttke 1990-, v2(C++) 2001-                                 *//
//* http://apps.jcns.fz-juelich.de/frida                                    *//
//***************************************************************************//

//! \file  mem.cpp
//! \brief NOlm: online memory and file selection

#include <algorithm>

#include "defs.hpp"
#include "olf.hpp"
#include "list.hpp"
#include "mem.hpp"


//***************************************************************************//
//*  Singleton instances                                                    *//
//***************************************************************************//

namespace NOlm {
    vector<POlo> MOM;
    vector<POlo> BAK;
    bool overwrite;
    CList FSel, FSelOld, FSelNew;
}


//***************************************************************************//
//*  File selection                                                         *//
//***************************************************************************//

namespace NOlm {

    //! Receive new file selection.
    void sel_push( CList* _FSel )
    {
        FSelOld = FSel;
        FSel = *_FSel;
    }
    
    //! Go back to previous file selection.
    void sel_pop()
    {
        FSel = FSelOld;
        FSelOld = CList();
    }

    //! Return true if file selection can be taken from 'inp'.
    bool sel_try( const string& inp )
    {
        try {
            CList newSel( inp, 0, MOM.size()-1 );
            NOlm::sel_push( &newSel );
            return true;
        } catch ( ... ) {
            return false;
        }
    }

    //! Start collection of new files.
    void sel_collect_begin()
    {   FSelNew.clear(); }

    //! End collection of new files, and make them the new default selection.
    void sel_collect_end()
    {
        if( !overwrite && FSelNew.size()>0 )
            sel_push( &FSelNew );
    }

    //! Step up or down.
    void sel_increment( int step )
    {
        for( int i=0; i<FSel.size(); ++i )
            FSel.V[i] += step;
    }

    //! Returns number of files in current selection.
    int sel_size()
    {   return FSel.size(); }

    //! Selected file numbers as string.
    string sel_str()
    {   return FSel.str(); }

    //! First selected file.
    POlo sel_first()
    {
        if( !FSel.size() )
            throw "empty file selection";
        return MOM[FSel.V[0]];
    }

    //! Maximum number of scans in selected files.
    int nJ_max()
    {
        int ret = 0;
        for( int iv: FSel.V )
            ret = max( ret, MOM[iv]->nJ() );
        return ret;
    }
}


//***************************************************************************//
//*  Iterators: Implementation                                              *//
//***************************************************************************//

namespace NOlm {

    //! Iterate over file selection. Implementation for use in IterateO.

    class MyIterate {
    public: // TODO: should become private
        int iV;
        CList Sel;
    public:
        MyIterate( int mode );
        void reset();
        POlo get();
        bool going() const;
        int k() const;
        int iteration() const;
        int size() const;
    };

    MyIterate::MyIterate( int mode )
    {
        if ( mode & IterALL )
            Sel = CList( 0,MOM.size()-1 );
        else {
            if ( !FSel.all_between( 0,MOM.size()-1 ) )
                throw "invalid file selection";
            Sel = FSel;
        }
        reset();
    }

    void MyIterate::reset()
    {   iV = 0; }

    POlo MyIterate::get()
    { 
        int k=Sel.V[iV];
        if( k<0 )
            throw "BUG: k<0 in file selection";
        if( k>=MOM.size() )
            throw "FileSelection["+S(iV)+"]="+S(k)+
                " exceeds online memory size "+S(MOM.size());
        ++iV;
        return MOM[k];
    }

    bool MyIterate::going() const
    {   return iV<Sel.size(); }

    int MyIterate::k() const
    {   return Sel.V[iV-1]; }

    int MyIterate::iteration() const
    {   return iV-1; }

    int MyIterate::size() const
    {   return Sel.V.size(); }
}


//***************************************************************************//
//*  Iterators: Interface                                                   *//
//***************************************************************************//

namespace NOlm {

    IterateO::IterateO( int mode ) :
        pimpl( new MyIterate( mode ) )
    {
        if ( !(mode & IterTOL) && !size() )
            throw "empty file selection";
    }

    IterateO::~IterateO()
    {}

    POlo IterateO::operator()()
    {   return pimpl->going() ? pimpl->get() : POlo(); }

    void IterateO::reset()
    {   return pimpl->reset(); }

    int IterateO::k() const
    {   return pimpl->k(); }

    int IterateO::iteration() const
    {   return pimpl->iteration(); }

    int IterateO::size() const
    {   return pimpl->size(); }

    POld IterateD::operator()()
    {
        POld e;
        while ( pimpl->going() && !(e = P2D( pimpl->get() )) ) {};
        return e;
    }

    int IterateD::size() const
    {
        IterateD iter2;
        int ret = 0;
        for( ; iter2(); ++ret ) {};
        return ret;
    }

    POlc IterateC::operator()()
    {
        POlc e;
        while ( pimpl->going() && !(e = P2C( pimpl->get() )) ) {};
        return e;
    }

    int IterateC::size() const
    {
        IterateC iter2;
        int ret = 0;
        for( ; iter2(); ++ret ) {};
        return ret;
    }
}


//***************************************************************************//
//*  Read access to online memory                                           *//
//***************************************************************************//

namespace NOlm {

    //! Return pointer to file with index k.
    POlo mem_get( int k )
    {
        if ( k<0 )
            throw "negative number " + S(k) + " cannot be a file index";
        if ( k>=MOM.size() )
            throw "file index " + S(k) + " not in online memory";
        return MOM[k];
    }

    //! Return pointer to file with index k, provided it contains data.
    POld mem_get_D( int k )
    {
        POld ret = P2D( mem_get( k ) );
        if( !ret )
            throw "file " + S(k) + " is not a data file";
        return ret;
    }

    //! Return pointer to file with index k, provided it contains curves.
    POlc mem_get_C( int k )
    {
        POlc ret = P2C( mem_get( k ) );
        if( !ret )
            throw "file " + S(k) + " is not a curve file";
        return ret;
    }
}

//***************************************************************************//
//*  Modify online memory, and adapt file selection                         *//
//***************************************************************************//

namespace NOlm {

    //! Store file f in online memory. If overwrite and k!=-1, overwrite file k.
    void mem_store( POlo f, int k )
    {
        if( overwrite && k!=-1 ){
            // replace file k
            if( k>=MOM.size() )
                throw "BUG: cannot overwrite inexistent file";
            MOM[k] = f;
        } else {
            // append at end of online memory
            MOM.push_back( f );
            FSelNew.append( MOM.size()-1 );
        }
    }

    void mem_store( POld fd, int k ) { mem_store( (POlo)fd, k ); }
    void mem_store( POlc fc, int k ) { mem_store( (POlo)fc, k ); }

    //! Remove selected files from online memory.
    void mem_del()
    {
        for (int i = FSel.size()-1; i>=0; --i)
            mem_del( FSel.V[i] );
        sel_pop();
    }

    //! Remove file k from online memory.
    void mem_del( int k )
    {
        MOM.erase(MOM.begin()+k);
        for( int i=0; i<FSelOld.size(); ++i )
            if ( FSelOld.V[i] > k )
                --FSelOld.V[i];
    }

    //! Append copy of selected files to online memory.
    void mem_copy()
    {
        IterateO fiter;
        while ( POlo fin=fiter() ){
            POlo fout( fin->new_olo( false ) );
            fout->copy_mainvec( fin );
            mem_store( fout );
        }
    }
    
    //! Move files to end of online memory.
    void mem_xchange()
    {
        mem_copy();
        mem_del();
    }
}

//***************************************************************************//
//*  Backup and retrieve                                                    *//
//***************************************************************************//

namespace NOlm {

    //! Copy selected files to backup.

    void mem_backup_store()
    {
        BAK.clear();
        IterateO fiter;
        while ( POlo fin=fiter() ){
            POlo fout( fin->new_olo( false ) );
            fout->copy_mainvec( fin );
            BAK.push_back( fout );
        }
    }

    //! Retrieve files from backup.

    void mem_backup_retrieve()
    {
        if( !BAK.size() )
            throw "no files in backup";
        sel_collect_begin();
        for( auto b: BAK )
            mem_store( b, -1 );
        sel_collect_end();
        BAK.clear();
    }
}
