//***************************************************************************//
//* FRIDA: fast reliable interactive data analysis                          *//
//* (C) Joachim Wuttke 1990-, v2(C++) 2001-                                 *//
//* http://apps.jcns.fz-juelich.de/frida                                    *//
//***************************************************************************//

//! \file  opr.cpp
//! \brief NOperate: functional operations.


#include <iostream>
#include <boost/format.hpp>

#include <readplus/ask.hpp>
#include <readplus/readln.hpp>

#include "defs.hpp"
#include "olf.hpp"
#include "mem.hpp"
#include "slice.hpp"
#include "reg.hpp"
#include "var.hpp"
#include "expr.hpp"
#include "opr.hpp"
#include "xax_lex.hpp"

using boost::format;


//***************************************************************************//
//* operate via generic expressions                                         *//
//***************************************************************************//


//! Display data. Use boolean expression to display only a subset of the data.

void NOperate::show( const string& subcmd ) 
{
    NOlm::IterateD fiter;

    static string expr;
    expr = sask( "Show if", expr);
    if (expr=="") expr = "1";
    PNode T = user_xaxparse( expr.c_str() );
    if( T->has_dummy() )
        throw "dummy argument t not allowed when operating on data";

    while( POld fin=fiter() ) {
        int k = fiter.k();
        if ( fiter.size()>1 )
            printf( ".. file %d:\n", fiter.k() );

        double lval;
        string outj, outp;

        // determine length of longest spectrum:
        int ni = 0;
        for ( int j=0; j<fin->nJ(); j++)
            ni = max( ni, fin->VS(j)->size() );

        // printout by i:
        for ( int i=0; i<ni; i++) {
            outj = "";
            outp = "";

            for( int j=0; j<fin->nJ(); j++ ){
                const PSpec s = fin->VS(j);
                if( i>= s->size() ) continue;
                T->tree_point_val( &lval, 0, k, j, i );
                if( !lval ) continue;
                outj += S(j) + " ";
                if      ( subcmd=="xyd" )
                    if( s->dy.size() )
                        outp += str( format( "%- 12.5g %- 12.5g +- %- 12.5g" )
                                     % s->x[i] % s->y[i] % s->dy[i] );
                    else
                        outp += str( format( "%- 12.5g %- 12.5g" )
                                     % s->x[i] % s->y[i] );
                else if ( subcmd=="y" )
                    outp += str( format( "%- 12.5g" ) % s->y[i] );
                else
                    throw "invalid subcmd " + subcmd;
                outp += " ";
            }

            if( outp=="" ) continue;

            printf( "%5u [%s] %s\n", i, outj.substr(0,outj.size()-1).c_str(), 
                    outp.substr(0,outp.size()-1).c_str() );
        }
    }
}


//! Delete or retain data points according to a boolean expression.

void NOperate::select( bool sel_del )
{
    NOlm::IterateO fiter;

    static string expr;
    expr = sask((sel_del ? "Delete" : "Retain"), expr);
    if (expr=="") return;
    PNode T = user_xaxparse( expr.c_str() );
    if( T->has_dummy() )
        throw "dummy argument t not allowed when operating on data";

    while ( POlo fin=fiter() ) {
        POld fd = P2D( fin );
        POlc fc = P2C( fin );

        POlo fout( fin->new_olo() );
        fout->lDoc.push_back((sel_del ? "md " : "mr ") + expr);

        for ( int j=0; j<fin->nJ(); j++ ) {
            if ( fd ) {
                PSpec ein = fd->VS(j);
                PSpec eout = PSpec( new CSpec );
                eout->copy_z_base( ein );
                vector<double> lvec( ein->size() );
                T->tree_vec_val( &lvec, 0, fiter.k(), j );
                bool with_dy = ein->dy.size();
                for ( int i=0; i<ein->size(); i++ ) {
                    if ( sel_del ^ (bool) lvec[i] ) {
                        if( with_dy )
                            eout->push_xyd( ein->x[i], ein->y[i], ein->dy[i] );
                        else
                            eout->push_xy( ein->x[i], ein->y[i] );
                    }
                }
                if ( eout->size() )
                    fout->V.push_back( eout );
            } else if ( fc ) {
                PCurve eout = PCurve( new CCurve( *(fc->VC(j)) ) );
                double lval;
                T->tree_point_val( &lval, 0, fiter.k(), j );
                if ( sel_del ^ (bool) lval )
                    fout->V.push_back( eout );
            }                
        }
        if( !fout->nJ() )
            throw "no data point left";
        NOlm::mem_store( fout, fiter.k() );
    }
}


//! Pointwise operation on x, y, or z<n>, according to arithmetic expression.

void NOperate::Pointwise( string llabel ) 
{
    NOlm::IterateO fiter;

    CVar lref(llabel);

    string expr = sask("Expression for " + llabel + " ?");
    if (expr=="") return;
    PNode T = user_xaxparse( expr.c_str() );
    // cout << "DEBUG: " << T->tree_info() << "\n";
    if( T->has_dummy() )
        throw "dummy argument t not allowed when operating on data";
 
    while( POlo fin=fiter() ){
        int k = fiter.k();
        POld fd = P2D( fin );
        POlc fc = P2C( fin );

        if ( lref.pointwise() && !fd )
            throw "no pointwise operation on curve";
        if( lref.typ==CVar::_CP && !fc )
            continue;

        POlo fout( fin->new_olo() );

        if( lref.typ!=CVar::_CP )
            fout->lDoc.push_back( "o" + lref.var_info() + " " + expr);
        
        CCoord co;
        T->set_coord( co, k );
        if        ( lref.typ==CVar::_X ) {
            fout->xco = co;
        } else if ( lref.typ==CVar::_Y ) {
            fout->yco = co;
        } else if ( lref.typ==CVar::_Z ) {
            if( lref.num==-1 )
                fout->ZCo.push_back( co );
            else if( lref.num >= fin->ZCo.size() )
                throw "no such z coordinate to operate on";
            else
                fout->ZCo[lref.num] = co;
        } else if ( lref.typ==CVar::_R ) {
            if( lref.num==-1 )
                fout->RPar.push_back( CParam( co ) );
            else if( lref.num >= fin->RPar.size() )
                throw "no such real par to operate on";
            else
                fout->RPar[lref.num].Co = co;
        } else if ( lref.typ==CVar::_CP ) {
            if( !fc )
                continue; // just ignore op operations on data files
            if( lref.num>=fc->nP )
                throw "no such parameter to operate on";
            NOlm::overwrite = 1;
        }

        if ( lref.typ==CVar::_R ) {
            fout->copy_mainvec( fin );
            double val;
            T->tree_point_val( &val, 0, k, 0 );
            if( lref.num==-1 )
                fout->RPar.back().val = val;
            else
                fout->RPar[lref.num].val = val;
        } else {
            fout->V.resize( fin->nJ() );
            for( int j=0; j<fin->nJ(); j++ ){
                fout->V[j] = fin->new_slice( j );
                if ( fin->V[j]->frozen )
                    continue;
                if ( lref.pointwise() ) {
                    PSpec eout( new CSpec( *(fd->VS(j)) ) );
                    vector<double> vout( fd->nPts(j) );
                    if ( lref.typ==CVar::_X || lref.typ==CVar::_DY ) {
                        T->tree_vec_val( &vout, 0, k, j );
                        if      ( lref.typ==CVar::_X )
                            (P2D(fout))->VS(j)->x = vout;
                        else if ( lref.typ==CVar::_DY )
                            (P2D(fout))->VS(j)->dy = vout;
                        else
                            throw "BUG: unexpected lref";
                    } else if ( lref.typ==CVar::_Y ) {
                        vector<double> dvout;
                        T->tree_vec_val( &vout, &dvout, k, j );
                        (P2D(fout))->VS(j)->y = vout;
                        if( dvout.size() )
                            (P2D(fout))->VS(j)->dy = dvout;
                        else
                            (P2D(fout))->VS(j)->dy.clear();
                    } else
                        throw "BUG: unexpected lref";
                } else if ( lref.typ==CVar::_Z ) {
                    double val;
                    T->tree_point_val( &val, 0, k, j );
                    if ( lref.num==-1 )
                        fout->V[j]->z.push_back( val );
                    else
                        fout->V[j]->z[lref.num] = val;
                } else if ( lref.typ==CVar::_CP ) {
                    double val;
                    T->tree_point_val( &val, 0, k, j );
                    (P2C(fout))->VC(j)->P[lref.num] = val;
                } else {
                    throw "BUG: invalid oxy";
                }
            }
        }
        NOlm::mem_store( fout, fiter.k() );
    }
}


//! Integral operation, using arithmetic expression.

void NOperate::Integral(void) 
{
    NOlm::IterateO fiter;

    string expr = sask( "Functional ?" );
    if ( expr=="" ) return;
    PNode T = user_xaxparse( expr.c_str() );
    if( T->has_dummy() )
        throw "dummy argument t not allowed when operating on data";

    double res;
    while( POlo fin=fiter() ) {
        int k = fiter.k();
        int nz = fin->ZCo.size();
        bool savable = nz > 0;
        
        POlo fout;
        PSpec sout;
        if ( savable ) {
            fout = POlo( fin->new_olo() );
            fout->lDoc.push_back( "oi " +expr );
            fout->ZCo.pop_back();
            fout->xco = fin->ZCo.back();
            T->set_coord( fout->yco, k );
            sout = PSpec( new CSpec );
            sout->z = fin->V[0]->z;
            sout->z.pop_back();
        }

        for ( int j=0; j<fin->nJ(); j++ ) {
            double xval, zval, dres;
            T->tree_point_val( &res, &dres, k, j );
            if (!savable) {
                cout << "result: " << res;
                if ( dres )
                    cout << " +- " << dres;
                cout << "\n";
                continue;
            }
            xval = fin->V[j]->z[nz-1];
            sout->push_xyd( xval, res, dres );
            if (nz>=2) { // new spectrum if jump in other z values
                zval = fin->V[j]->z[nz-2];
                if ( j+1 < fin->nJ() && fin->V[j+1]->z[nz-2] != zval ) {
                    fout->V.push_back( sout );
                    sout = PSpec( new CSpec );
                    sout->z = fin->V[j+1]->z;
                    sout->z.pop_back();
                }
            }
        }
        if ( savable ) {
            sout->purge_dy();
            fout->V.push_back( sout );
            NOlm::mem_store( fout, fiter.k() );
        }
        if ( !savable )
            NReg::lastresult = res;
    }
}


//! Integral operations to generate not only new y, but also new x.

void NOperate::IntXY( string mode ) 
{
    NOlm::IterateD fiter;

    bool with_d;
    if      ( mode=="xy" )
        with_d = false;
    else if ( mode=="xyd" )
        with_d = true;
    else
        throw "invalid command";

    static int icolx = 0, icoly = 1, icold = 2;
    static CCoord xco, yco;
    icolx = iask( "x from column", icolx );
    if( icolx<0) return;
    icoly = iask( "y from column", icoly );
    if( icoly<0) return;
    if ( with_d ) {
        icold = iask( "dy from column", icold );
        if( icold<0) return;
    }
    xco = CCoord( "x", "" );
    yco = CCoord( "y", "" );

    while( POld fin=fiter() ) {
        int nz = fin->ZCo.size();

        POld fout( fin->new_old() );
        string lin = S("oi") + mode + " " + S(icolx) + " " + S(icoly);
        if ( with_d )
            lin += " " + S(icold);
        fout->lDoc.push_back( lin );

        fout->xco = xco;
        fout->yco = yco;

        PSpec sout( new CSpec );

        for ( int j=0; j<fin->nJ(); j++ ) {
            int n = fin->nPts(j);
            if( icolx>=n || icoly>=n || (with_d && icold>=n) )
                throw "not enough columns in spectrum " + S(j);
            if ( with_d )
                sout->push_xyd( fin->VS(j)->y[icolx],
                                fin->VS(j)->y[icoly],
                                fin->VS(j)->y[icold] );
            else
                sout->push_xy( fin->VS(j)->y[icolx],
                               fin->VS(j)->y[icoly] );
            sout->z = fin->VS(j)->z;
            if ( nz ) { // new spectrum if jump in any z values
                if ( j+1 < fin->nJ() && fin->V[j+1]->z != fin->V[j]->z ) {
                    fout->V.push_back( sout );
                    sout = PSpec( new CSpec );
                }
            }
        }
        fout->V.push_back( sout );
        NOlm::mem_store( fout, fiter.k() );
    }
}


//***************************************************************************//
//* functional operations                                                   *//
//***************************************************************************//

//! Functional integration, derivation, ...

void NOperate::Functional( const string& subcmd )
{
    NOlm::IterateD fiter;
    while ( POld fin=fiter() ) {
        POld fout( fin->new_old() );
        fout->lDoc.push_back( "of"+subcmd );

        if        ( subcmd=="i" ) { // integrate
            fout->yco.name = "Int d" + fin->xco.name + " " + fin->yco.name;
            fout->yco.unit = fin->xco.unit + " * " + fin->yco.unit;
        } else if ( subcmd=="s" ) { // sum
            fout->yco.name = "Sum_0^i " + fin->yco.name;
            fout->yco.unit = fin->yco.unit;
        } else if ( subcmd=="d" ) { // derive
            fout->yco.name = "d" + fin->yco.name + " / d" + fin->xco.name;
            fout->yco.unit = fin->yco.unit + " / " + fin->xco.unit;
        } else if ( subcmd=="ac" ) { // autocorrelate
            fout->yco.name = "ac{" + fin->yco.name + "}";
            fout->yco.unit = fin->yco.unit + "^2";
        } else {
            throw "invalid 'of' command";
        }

        for ( int j=0; j<fin->nJ(); j++ ) {
            PSpec sin = fin->VS(j);
            PSpec sout( new CSpec );
            sout->copy_z_base( sin );
            int n = sin->size();

            if        ( subcmd=="i" ) { // integrate
                sout->x = sin->x;
                sout->y.resize( n );
                sout->y[0] = 0;
                for ( int i=1; i<n; ++i )
                    sout->y[i] = sout->y[i-1] + 
                        (sin->x[i] - sin->x[i-1]) * 
                        (sin->y[i] + sin->y[i-1]) / 2;
            } else if ( subcmd=="s" ) { // sum
                sout->x = sin->x;
                sout->y.resize( n );
                sout->y[0] = sin->y[0];
                for ( int i=1; i<n; ++i )
                    sout->y[i] = sout->y[i-1] + sin->y[i];
            } else if ( subcmd=="d" ) { // derive
                for( int i=0; i<n-1; ++i ){
                    double dx = sin->x[i+1] - sin->x[i];
                    sout->push_xy(
                        (sin->x[i+1] + sin->x[i] ) / 2,
                        dx==0 ? 0 : (sin->y[i+1] - sin->y[i]) / dx );
                }
            } else if ( subcmd=="ac" ) { // autocorrelate
                throw "currently disabled, old code in"
                    " old/src-fragm/opr-func-ac.cpp";
            }
            fout->V.push_back( sout );
        }
        NOlm::mem_store( fout, fiter.k() );
    }
}
