//**************************************************************************//
//* FRIDA: fast reliable interactive data analysis                         *//
//* (C) Joachim Wuttke 1990-, C++ rewrite 2001-                            *//
//* http://apps.jcns.fz-juelich.de/frida                                   *//
//**************************************************************************//

//! \file toplevel.cpp
//! \brief Execution context CFrida.


//! \mainpage FRIDA: fast reliable interactive data analysis 
 
//! \par Homepage:
//!           http://apps.jcns.fz-juelich.de/frida

//! Useful entry points in this source documentation:
//! - initializations, main loop: 
//!     <a href="frida2_8cpp_source.html">frida2.cpp</a>
//! - commands:
//!     <a href="commands_8cpp_source.html">commands.cpp</a>
//! - online memory:
//!     <a href="namespaceNOlm.html">namespace NOlm</a>
//! - data model root:
//!     <a href="classCOlo.html">class COlo</a>
//!
//! Used libraries:
//! - Minor utilities:
//!     <a href="../frida-libs/trivia/index.html">libtrivia</a>
//! - Dialogue interface:
//!     <a href="../frida-libs/readplus/index.html">libreadplus</a>


#include <cstdlib>
#include <iostream>
#include <string>
#include <vector>
#include <gsl/gsl_errno.h>

#include <trivia/string_ops.hpp>
#include <readplus/ask.hpp>
#include <readplus/readln.hpp>

#include "toplevel.hpp"
#include "config.hpp"
#include "defs.hpp"
#include "olf.hpp"
#include "mem.hpp"
#include "func.hpp"
#include "geni.hpp"
#include "integrate.hpp"
#include "dualplot.hpp"
#include "commands.hpp"
#include "calc.hpp"

using namespace std;


//! Error handler for C code from the GNU Scientific Library.

void my_gsl_error_handler (
    const char * reason, const char * file, int line, int gsl_errno )
{
    cout << "GSL error " << gsl_errno << ": " << reason << " in " <<
        file << ", line " << line << "\n";
    // throw does not work (probably because GSL is in C, not in C++)
}


//! Prepare interactive Frida session.

CFrida::CFrida()
{
    // External setup file:
#ifdef CONFDIR
    try {
        NRead::exemac("\\i " CONFDIR "/frida.ini");
    } catch ( string& s ) {
        cerr << s << endl;
    }
#endif

    // Math:
    gsl_set_error_handler( my_gsl_error_handler );
    NFunctions::initialize();
    NGeni::initialize();
    NCvin::initialize();

    // Plot windows:
    NPloWin::initialize();

    // time monitor:
    clock_start = clock();
    clock_cmd_i = clock();
    time_cmd=0;
}


//! Prompts user for a command, and returns it.

string CFrida::prompt()
{
    try{
        return wask( NOlm::sel_str() + " > " );
    } catch( string& s ) {
        cerr << s << endl;
        NRead::clear();
    } catch( const char* s ) {
        cerr << s << endl;
        NRead::clear();
    } catch( ... ) {
        cerr << "catched invalid exception\n";
        NRead::clear();
    }
    return "";
}


//! CExecutes one Frida command.

void CFrida::execute( const string cmd )
{
    try{
        // Empty input, comment lines.
        if ( cmd=="" || cmd[0]=='#' )
            return;
        
        // Is it a list?
        if ( cmd.find_first_of("-*0123456789")==0 ) {
            if( NOlm::sel_try( cmd ) )
                return; // yes, it was a list
        }

        // Is it a command?
        clock_cmd_i = clock();
        if (cmd == "t") {
            double time_total =
                ((double) (clock() - clock_start)) / CLOCKS_PER_SEC;
            printf( "total frida CPU time now %g\n", time_total );
            printf( "last command took %g\n", time_cmd );
            return;
        } else if ( fridaCommand( cmd ) ) {
            return;
        }

        // Is it a expression? If not, throw an error.
        NCalc::Calculator( cmd );

    } catch( string& s ) {
        cerr << s << endl;
        NRead::clear();
    } catch( const char* s ) {
        cerr << s << endl;
        NRead::clear();
    } catch( ... ) {
        cerr << "catched invalid exception\n";
        NRead::clear();
    }
    time_cmd = ((double) (clock() - clock_cmd_i)) / CLOCKS_PER_SEC;
}
